@@ -55,8 +55,8 @@ require 'diakonos/readline'
 
 module Diakonos
 
-    VERSION = '0.8.3'
-    LAST_MODIFIED = 'February 7, 2007'
+    VERSION = '0.8.4'
+    LAST_MODIFIED = 'May 23, 2008'
 
     DONT_ADJUST_ROW = false
     ADJUST_ROW = true
@@ -84,15 +84,17 @@ module Diakonos
     CHOICE_CANCEL = 3
     CHOICE_YES_TO_ALL = 4
     CHOICE_NO_TO_ALL = 5
+    CHOICE_YES_AND_STOP = 6
     CHOICE_KEYS = [
         [ ?n, ?N ],
         [ ?y, ?Y ],
         [ ?a, ?A ],
         [ ?c, ?C, ESCAPE, CTRL_C, CTRL_D, CTRL_Q ],
         [ ?e ],
-        [ ?o ]
+        [ ?o ],
+        [ ?s ],
     ]
-    CHOICE_STRINGS = [ '(n)o', '(y)es', '(a)ll', '(c)ancel', 'y(e)s to all', 'n(o) to all' ]
+    CHOICE_STRINGS = [ '(n)o', '(y)es', '(a)ll', '(c)ancel', 'y(e)s to all', 'n(o) to all', 'yes and (s)top' ]
 
     BOL_ZERO = 0
     BOL_FIRST_CHAR = 1
@@ -116,6 +118,7 @@ module Diakonos
         'clearMatches',
         'closeFile',
         'collapseWhitespace',
+        'close_code',
         'copySelection',
         'cursorBOF',
         'cursorBOL',
@@ -149,6 +152,8 @@ module Diakonos
         'indent',
         'insertSpaces',
         'insertTab',
+        'joinLines',
+        'list_buffers',
         'loadConfiguration',
         'loadScript',
         'newFile',
@@ -173,6 +178,7 @@ module Diakonos
         'revert',
         'saveFile',
         'saveFileAs',
+        'select_block',
         'scrollDown',
         'scrollUp',
         'searchAndReplace',
@@ -200,7 +206,7 @@ module Diakonos
 class Diakonos
     attr_reader :win_main, :settings, :token_regexps, :close_token_regexps,
         :token_formats, :diakonos_home, :script_dir, :diakonos_conf, :display_mutex,
-        :indenters, :unindenters, :clipboard, :do_display,
+        :indenters, :unindenters, :closers, :clipboard, :do_display,
         :current_buffer, :list_filename, :hooks, :last_commands, :there_was_non_movement
 
 
@@ -268,7 +274,7 @@ class Diakonos
         while argv.length > 0
             arg = argv.shift
             case arg
-                when '--help'
+                when '-h', '--help'
                     printUsage
                     exit 1
                 when '-ro'
@@ -372,26 +378,26 @@ class Diakonos
         end
     end
     
-    def fetch_conf( location = "tags/v#{VERSION}" )
-        require 'open-uri'
-        found = false
-        puts "Fetching configuration from #{location}..."
-        
-        begin
-            open( "http://rome.purepistos.net/issues/diakonos/browser/#{location}/diakonos.conf?format=raw" ) do |http|
-                text = http.read
-                if Regexp.new( "No node /#{location}/diakonos.conf" ) !~ text
-                    found = true
-                    File.open( @diakonos_conf, 'w' ) do |f|
-                        f.puts text
-                    end
-                end
+    def fetch_conf( location = "v#{VERSION}" )
+      require 'open-uri'
+      found = false
+      puts "Fetching configuration from #{location}..."
+      
+      begin
+        open( "http://github.com/Pistos/diakonos/tree/#{location}/diakonos.conf?raw=true" ) do |http|
+          text = http.read
+          if text =~ /key/ and text =~ /colour/ and text =~ /lang/
+            found = true
+            File.open( @diakonos_conf, 'w' ) do |f|
+              f.puts text
             end
-        rescue OpenURI::HTTPError => e
-            $stderr.puts "Failed to fetch from #{location}."
+          end
         end
-            
-        return found
+      rescue OpenURI::HTTPError => e
+        $stderr.puts "Failed to fetch from #{location}."
+      end
+      
+      return found
     end
     
     def loadConfiguration
@@ -429,7 +435,7 @@ class Diakonos
             case answer
                 when /^y/i
                     if not fetch_conf
-                        fetch_conf 'trunk'
+                        fetch_conf 'master'
                     end
             end
             
@@ -448,6 +454,7 @@ class Diakonos
         @unindenters = Hash.new
         @filemasks = Hash.new
         @bangmasks = Hash.new
+        @closers = Hash.new
 
         @settings = Hash.new
         # Setup some defaults
@@ -573,9 +580,27 @@ class Diakonos
                     @filemasks[ $1 ] = Regexp.new arg
                 when /^lang\.(.+?)\.bangmask$/
                     @bangmasks[ $1 ] = Regexp.new arg
+                when /^lang\.(.+?)\.closers\.(.+?)\.(.+?)$/
+                    @closers[ $1 ] ||= Hash.new
+                    @closers[ $1 ][ $2 ] ||= Hash.new
+                    @closers[ $1 ][ $2 ][ $3.to_sym ] = case $3
+                        when 'regexp'
+                            Regexp.new arg
+                        when 'closer'
+                            begin
+                                eval( "Proc.new " + arg )
+                            rescue Exception => e
+                                showException(
+                                    e,
+                                    [
+                                        "Failed to process Proc for #{command}.",
+                                    ]
+                                )
+                            end
+                    end
                 when "context.visible", "context.combined", "eof_newline", "view.nonfilelines.visible",
                         /^lang\.(.+?)\.indent\.(?:auto|roundup|using_tabs)$/,
-                        "found_cursor_start", "convert_tabs"
+                        "found_cursor_start", "convert_tabs", 'delete_newline_on_delete_to_eol'
                     @settings[ command ] = arg.to_b
                 when "context.format", "context.separator.format", "status.format"
                     @settings[ command ] = arg.toFormatting
@@ -697,7 +722,7 @@ class Diakonos
             hook.sort { |a,b| a[ :priority ] <=> b[ :priority ] }
         end
 
-        setILine "Diakonos #{VERSION} (#{LAST_MODIFIED})   F1 for help  F12 to configure"
+        setILine "Diakonos #{VERSION} (#{LAST_MODIFIED})   F1 for help  F12 to configure   Ctrl-Q to quit"
         
         num_opened = 0
         if @files.length == 0 and @read_only_files.length == 0
@@ -770,7 +795,6 @@ class Diakonos
         
             if context.empty?
                 if c > 31 and c < 255 and c != BACKSPACE
-                    debugLog "char: #{c}"
                     if @macro_history != nil
                         @macro_history.push "typeCharacter #{c}"
                     end
@@ -837,7 +861,6 @@ class Diakonos
         clip_filename = @diakonos_home + "/clips.txt"
         File.open( clip_filename, "w" ) do |f|
             @clipboard.each do |clip|
-                log clip
                 f.puts clip
                 f.puts "---------------------------"
             end
@@ -1053,7 +1076,7 @@ class Diakonos
                 f.puts
                 f.puts "----------------------------------------------------"
                 f.puts "If you can reproduce this error, please report it at"
-                f.puts "http://rome.purepistos.net/issues/diakonos/newticket !"
+                f.puts "http://linis.purepistos.net/ticket/list/Diakonos !"
                 f.puts "----------------------------------------------------"
                 f.puts e.backtrace
             end
@@ -1329,6 +1352,34 @@ class Diakonos
     
     def closeListBuffer
         closeFile( @list_buffer )
+        @list_buffer = nil
+    end
+    def showing_list?
+      @list_buffer
+    end
+    def current_list_item
+      if @list_buffer
+        @list_buffer.select_current_line
+      end
+    end
+    def select_list_item
+      if @list_buffer
+        line = @list_buffer.select_current_line
+        @list_buffer.display
+        line
+      end
+    end
+    def previous_list_item
+      if @list_buffer
+        cursorUp
+        @list_buffer[ @list_buffer.currentRow ]
+      end
+    end
+    def next_list_item
+      if @list_buffer
+        cursorDown
+        @list_buffer[ @list_buffer.currentRow ]
+      end
     end
     
     def runHookProcs( hook_id, *args )
@@ -1396,6 +1447,10 @@ class Diakonos
     def clearMatches
         @current_buffer.clearMatches Buffer::DO_DISPLAY
     end
+    
+    def close_code
+        @current_buffer.close_code
+    end
 
     # Returns the choice the user made, or nil if the user was not prompted to choose.
     def closeFile( buffer = @current_buffer, to_all = nil )
@@ -1806,11 +1861,11 @@ class Diakonos
             else
                 end_row = mark.end_row - 1
             end
-            (mark.start_row...end_row).each do |row|
+            (mark.start_row..end_row).each do |row|
                 @current_buffer.indent row, Buffer::DONT_DISPLAY
             end
             @do_display = true
-            @current_buffer.indent( end_row ) 
+            @current_buffer.display
         else
             @current_buffer.indent
         end
@@ -1827,7 +1882,23 @@ class Diakonos
     def insertTab
         typeCharacter( TAB )
     end
+    
+    def joinLines
+        @current_buffer.joinLines( @current_buffer.currentRow, Buffer::STRIP_LINE )
+    end
 
+    def list_buffers
+      File.open( @list_filename, "w" ) do |f|
+        f.puts @buffers.keys.map { |name| "#{name}\n" }.sort
+      end
+      openListBuffer
+      filename = getUserInput( "Switch to buffer: " )
+      buffer = @buffers[ filename ]
+      if buffer
+        switchTo buffer
+      end
+    end
+    
     def loadScript( name_ = nil )
         if name_ == nil
             name = getUserInput( "File to load as script: ", @rlh_files )
@@ -2013,11 +2084,11 @@ class Diakonos
         if( @current_buffer.changing_selection )
             @do_display = false
             mark = @current_buffer.selection_mark
-            (mark.start_row...mark.end_row).each do |row|
+            (mark.start_row..mark.end_row).each do |row|
                 @current_buffer.parsedIndent row, Buffer::DONT_DISPLAY
             end
             @do_display = true
-            @current_buffer.parsedIndent mark.end_row
+            @current_buffer.display
         else
             @current_buffer.parsedIndent
         end
@@ -2151,6 +2222,16 @@ class Diakonos
             #end
         end
     end
+    
+    def select_block( beginning = nil, ending = nil, including_ending = true )
+      if beginning.nil?
+        beginning = Regexp.new( getUserInput( "Start at regexp: " ) )
+      end
+      if ending.nil?
+        ending = Regexp.new( getUserInput( "End before regexp: " ) )
+      end
+      @current_buffer.select( beginning, ending, including_ending )
+    end
 
     def scrollDown
         @current_buffer.pitchView( @settings[ "view.scroll_amount" ] || 1 )
@@ -2374,11 +2455,11 @@ class Diakonos
             else
                 end_row = mark.end_row - 1
             end
-            (mark.start_row...end_row).each do |row|
+            (mark.start_row..end_row).each do |row|
                 @current_buffer.unindent row, Buffer::DONT_DISPLAY
             end
             @do_display = true
-            @current_buffer.unindent( end_row ) 
+            @current_buffer.display
         else
             @current_buffer.unindent
         end
