@@ -25,6 +25,8 @@ class Buffer
     DONT_PITCH_CURSOR = false
     CLEAR_STACK_POINTER = true
     DONT_CLEAR_STACK_POINTER = false
+    STRIP_LINE = true
+    DONT_STRIP_LINE = false
 
     # Set name to nil to create a buffer that is not associated with a file.
     def initialize( diakonos, name, read_only = false )
@@ -121,6 +123,7 @@ class Buffer
         @indenters = @diakonos.indenters[ @language ]
         @unindenters = @diakonos.unindenters[ @language ]
         @preventers = @settings[ "lang.#{@language}.indent.preventers" ]
+        @closers = @diakonos.closers[ @language ] || Hash.new
         @auto_indent = @settings[ "lang.#{@language}.indent.auto" ]
         @indent_size = ( @settings[ "lang.#{@language}.indent.size" ] or 4 )
         @indent_roundup = ( @settings[ "lang.#{@language}.indent.roundup" ] or true )
@@ -209,7 +212,7 @@ class Buffer
                             linestr[ 0 ] = ( @settings[ "view.nonfilelines.character" ] or "~" )
                         end
                         
-                        @win_main.addstr_ linestr
+                        @win_main.addstr linestr
                     end
                     
                     @win_main.setpos( @last_screen_y , @last_screen_x )
@@ -221,9 +224,9 @@ class Buffer
                     
                     Curses::curs_set 1
                 rescue Exception => e
-                    $diakonos.log( "Display Exception:" )
-                    $diakonos.log( e.message )
-                    $diakonos.log( e.backtrace.join( "\n" ) )
+                    @diakonos.log( "Display Exception:" )
+                    @diakonos.log( e.message )
+                    @diakonos.log( e.backtrace.join( "\n" ) )
                     showException e
                 end
                 @diakonos.display_mutex.unlock
@@ -343,6 +346,45 @@ class Buffer
     def selection_mark
         return @text_marks[ SELECTION ]
     end
+    
+    def select_current_line
+      @text_marks[ SELECTION ] = TextMark.new(
+        @last_row,
+        0,
+        @last_row,
+        @lines[ @last_row ].size,
+        @selection_formatting
+      )
+      @lines[ @last_row ]
+    end
+    
+    def select( from_regexp, to_regexp, include_ending = true )
+      start_row = nil
+      
+      @lines[ 0..@last_row ].reverse.each_with_index do |line,index|
+        if line =~ from_regexp
+          start_row = @last_row - index
+          break
+        end
+      end
+      if start_row
+        end_row = nil
+        @lines[ start_row..-1 ].each_with_index do |line,index|
+          if line =~ to_regexp
+            end_row = start_row + index
+            break
+          end
+        end
+        if end_row
+          if include_ending
+            end_row += 1
+          end
+          anchorSelection( start_row, 0, DONT_DISPLAY )
+          cursorTo( end_row, 0 )
+          display
+        end
+      end
+    end
 
     # Prints text to the screen, truncating where necessary.
     # Returns nil if the string is completely off-screen.
@@ -392,7 +434,7 @@ class Buffer
                 @win_main.attrset text_mark.formatting
                 if ( (text_mark.start_row + 1) .. (text_mark.end_row - 1) ) === row
                     @win_main.setpos( cury, curx )
-                    @win_main.addstr_ string
+                    @win_main.addstr string
                 elsif row == text_mark.start_row and row == text_mark.end_row
                     expanded_col = tabExpandedColumn( text_mark.start_col, row )
                     if expanded_col < @left_column + Curses::cols
@@ -400,7 +442,7 @@ class Buffer
                         right = tabExpandedColumn( text_mark.end_col, row ) - @left_column
                         if left < right
                             @win_main.setpos( cury, curx + left )
-                            @win_main.addstr_ string[ left...right ]
+                            @win_main.addstr string[ left...right ]
                         end
                     end
                 elsif row == text_mark.start_row
@@ -408,12 +450,12 @@ class Buffer
                     if expanded_col < @left_column + Curses::cols
                         left = [ expanded_col - @left_column, 0 ].max
                         @win_main.setpos( cury, curx + left )
-                        @win_main.addstr_ string[ left..-1 ]
+                        @win_main.addstr string[ left..-1 ]
                     end
                 elsif row == text_mark.end_row
                     right = tabExpandedColumn( text_mark.end_col, row ) - @left_column
                     @win_main.setpos( cury, curx )
-                    @win_main.addstr_ string[ 0...right ]
+                    @win_main.addstr string[ 0...right ]
                 else
                     # This row not in selection.
                 end
@@ -426,7 +468,7 @@ class Buffer
         return if string == nil
 
         @win_main.attrset formatting
-        @win_main.addstr_ string
+        @win_main.addstr string
     end
 
     # This method assumes that the cursor has been setup already at
@@ -644,10 +686,8 @@ class Buffer
                 if col == line.length
                     if row < @lines.length - 1
                         # Delete newline, and concat next line
-                        takeSnapshot( TYPING )
-                        @lines[ row ] << @lines.delete_at( row + 1 )
+                        joinLines( row )
                         cursorTo( @last_row, @last_col )
-                        setModified
                     end
                 else
                     takeSnapshot( TYPING )
@@ -658,6 +698,29 @@ class Buffer
         end
     end
     
+    def joinLines( row = @last_row, strip = DONT_STRIP_LINE )
+        takeSnapshot( TYPING )
+        next_line = @lines.delete_at( row + 1 )
+        if strip
+            next_line = ' ' + next_line.strip
+        end
+        @lines[ row ] << next_line
+        setModified
+    end
+    
+    def close_code
+        line = @lines[ @last_row ]
+        @closers.each_value do |h|
+            h[ :regexp ] =~ line
+            lm = Regexp.last_match
+            if lm
+                insertString h[ :closer ].call( lm ).to_s
+            else
+                @diakonos.log h[ :regexp ].inspect + " does not match '#{line}'"
+            end
+        end
+    end
+    
     def collapseWhitespace
         removeSelection( DONT_DISPLAY ) if selection_mark != nil
         
@@ -698,12 +761,19 @@ class Buffer
 
         row = @last_row
         col = @last_col
+        
         takeSnapshot
-        retval = [ @lines[ row ][ col..-1 ] ]
-        @lines[ row ] = @lines[ row ][ 0...col ]
+        if @settings[ 'delete_newline_on_delete_to_eol' ] and col == @lines[ row ].size
+          next_line = @lines.delete_at( row + 1 )
+          @lines[ row ] << next_line
+          retval = ''
+        else        
+          retval = [ @lines[ row ][ col..-1 ] ]
+          @lines[ row ] = @lines[ row ][ 0...col ]
+        end
         setModified
 
-        return retval
+        retval
     end
 
     def carriageReturn
@@ -1374,7 +1444,7 @@ class Buffer
             if replacement != nil
                 choice = @diakonos.getChoice(
                     "Replace?",
-                    [ CHOICE_YES, CHOICE_NO, CHOICE_ALL, CHOICE_CANCEL ],
+                    [ CHOICE_YES, CHOICE_NO, CHOICE_ALL, CHOICE_CANCEL, CHOICE_YES_AND_STOP ],
                     CHOICE_YES
                 )
                 case choice
@@ -1387,6 +1457,9 @@ class Buffer
                         find( regexps, direction, replacement )
                     when CHOICE_CANCEL
                         # Do nothing further.
+                    when CHOICE_YES_AND_STOP
+                        paste [ replacement ]
+                        # Do nothing further.
                 end
             end
         else
